==> MindsDB-23.11.1.0.tar.gz: MindsDB-23.11.1.0/mindsdb/integrations/handlers/rag_handler/ingest.py <==
import time

import pandas as pd
from langchain.schema import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import VectorStore

from mindsdb.integrations.handlers.rag_handler.settings import (
    PersistedVectorStoreSaver,
    PersistedVectorStoreSaverConfig,
    RAGBaseParameters,
    VectorStoreFactory,
    df_to_documents,
    get_chroma_client,
    load_embeddings_model,
    url_to_documents,
)
from mindsdb.utilities.log import get_log

logger = get_log(__name__)


def validate_document(doc) -> bool:
    """Check an individual document."""
    # Example checks
    if not isinstance(doc, Document):
        return False

    if not doc.page_content:
        return False

    return True


def validate_documents(documents) -> bool:
    """Validate document list format."""

    if not isinstance(documents, list):
        return False

    if not documents:
        return False

    # Check fields/format of a document
    return all([validate_document(doc) for doc in documents])


class RAGIngestor:
    """A class for converting a dataframe and/or url to a vectorstore embedded with a given embeddings model"""

    def __init__(
        self,
        args: RAGBaseParameters,
        df: pd.DataFrame,
    ):
        self.args = args
        self.df = df
        self.embeddings_model_name = args.embeddings_model_name

        self.vector_store = VectorStoreFactory.get_vectorstore_class(
            args.vector_store_name
        )

    def split_documents(self, chunk_size, chunk_overlap) -> list:
        # Load documents and split in chunks
        logger.info(f"Loading documents from input data")

        documents = []

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

        if self.df is not None:
            # if user provides a dataframe, load documents from dataframe
            documents.extend(
                df_to_documents(
                    df=self.df, page_content_columns=self.args.context_columns
                )
            )

        if self.args.url:
            # if user provides a url, load documents from url
            documents.extend(url_to_documents(self.args.url))

        # split documents into chunks of text
        texts = text_splitter.split_documents(documents)
        logger.info(f"Loaded {len(documents)} documents from input data")
        logger.info(f"Split into {len(texts)} chunks of text (tokens)")

        return texts

    def create_db_from_documents(self, documents, embeddings_model) -> VectorStore:
        """Create DB from documents."""

        if self.args.vector_store_name == "chroma":

            return self.vector_store.from_documents(
                documents=documents,
                embedding=embeddings_model,
                client=get_chroma_client(
                    persist_directory=self.args.vector_store_storage_path
                ),
                collection_name=self.args.collection_name,
            )
        else:
            return self.vector_store.from_documents(
                documents=documents,
                embedding=embeddings_model,
                index_name=self.args.collection_name,
            )

    def create_db_from_texts(self, documents, embeddings_model) -> VectorStore:
        """Create DB from text content."""

        texts = [doc.page_content for doc in documents]
        metadata = [doc.metadata for doc in documents]

        return self.vector_store.from_texts(
            texts=texts, embedding=embeddings_model, metadatas=metadata
        )

    def embeddings_to_vectordb(self) -> None:
        """Create vectorstore from documents and store locally."""

        start_time = time.time()

        # Load documents and splits in chunks (if not in evaluation_type mode)
        documents = self.split_documents(
            chunk_size=self.args.chunk_size, chunk_overlap=self.args.chunk_overlap
        )

        # Load embeddings model
        embeddings_model = load_embeddings_model(self.embeddings_model_name)

        logger.info(f"Creating vectorstore from documents")

        if not validate_documents(documents):
            raise ValueError("Invalid documents")

        try:
            db = self.create_db_from_documents(documents, embeddings_model)
        except Exception as e:
            logger.error(
                f"Error loading using 'from_documents' method, trying 'from_text': {e}"
            )
            try:
                db = self.create_db_from_texts(documents, embeddings_model)
                logger.info(f"successfully loaded using 'from_text' method: {e}")

            except Exception as e:
                logger.error(f"Error creating from texts: {e}")
                raise e

        config = PersistedVectorStoreSaverConfig(
            vector_store_name=self.args.vector_store_name,
            vector_store=db,
            persist_directory=self.args.vector_store_storage_path,
            collection_name=self.args.collection_name,
        )

        vector_store_saver = PersistedVectorStoreSaver(config)

        vector_store_saver.save_vector_store(db)

        db = None  # Free up memory
        end_time = time.time()
        elapsed_time = end_time - start_time

        logger.info(
            "Fished creating vectorstore from documents. It took: {elapsed_time/60} minutes"
        )

        logger.info("Finished creating vectorstore from documents.")
        logger.info(f"Elapsed time: {round(elapsed_time / 60)} minutes")

==> MindsDB-23.12.4.2.tar.gz: MindsDB-23.12.4.2/mindsdb/integrations/handlers/rag_handler/ingest.py <==
import time
from typing import List

import pandas as pd
from langchain.schema import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import VectorStore

from mindsdb.integrations.handlers.rag_handler.settings import (
    PersistedVectorStoreSaver,
    PersistedVectorStoreSaverConfig,
    RAGBaseParameters,
    VectorStoreFactory,
    df_to_documents,
    get_chroma_client,
    load_embeddings_model,
    url_to_documents,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def validate_document(doc) -> bool:
    """Check an individual document."""
    # Example checks
    if not isinstance(doc, Document):
        return False

    if not doc.page_content:
        return False

    return True


def validate_documents(documents) -> bool:
    """Validate document list format."""

    if not isinstance(documents, list):
        return False

    if not documents:
        return False

    # Check fields/format of a document
    return all([validate_document(doc) for doc in documents])


class RAGIngestor:
    """A class for converting a dataframe and/or url to a vectorstore embedded with a given embeddings model"""

    def __init__(
        self,
        args: RAGBaseParameters,
        df: pd.DataFrame,
    ):
        self.args = args
        self.df = df
        self.embeddings_model_name = args.embeddings_model_name

        self.vector_store = VectorStoreFactory.get_vectorstore_class(
            args.vector_store_name
        )

    def split_documents(self, chunk_size, chunk_overlap) -> list:
        # Load documents and split in chunks
        logger.info(f"Loading documents from input data")

        documents = []

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

        if self.df is not None:
            # if user provides a dataframe, load documents from dataframe
            documents.extend(
                df_to_documents(
                    df=self.df,
                    page_content_columns=self.args.context_columns,
                    url_column_name=self.args.url_column_name,
                )
            )

        if self.args.url:
            # if user provides a url, load documents from url
            documents.extend(url_to_documents(self.args.url))

        n_tokens = sum([len(doc.page_content) for doc in documents])

        # split documents into chunks of text
        texts = text_splitter.split_documents(documents)
        logger.info(f"Loaded {len(documents)} documents from input data")
        logger.info(f"Total number of tokens: {n_tokens}")
        logger.info(f"Split into {len(texts)} chunks of text (tokens)")

        return texts

    def create_db_from_documents(self, documents, embeddings_model) -> VectorStore:
        """Create DB from documents."""

        if self.args.vector_store_name == "chromadb":

            return self.vector_store.from_documents(
                documents=documents,
                embedding=embeddings_model,
                client=get_chroma_client(
                    persist_directory=self.args.vector_store_storage_path
                ),
                collection_name=self.args.collection_name,
            )
        else:
            return self.create_db_from_texts(documents, embeddings_model)

    def create_db_from_texts(self, documents, embeddings_model) -> VectorStore:
        """Create DB from text content."""

        texts = [doc.page_content for doc in documents]
        metadata = [doc.metadata for doc in documents]

        return self.vector_store.from_texts(
            texts=texts, embedding=embeddings_model, metadatas=metadata
        )

    @staticmethod
    def _create_batch_embeddings(documents: List[Document], embeddings_batch_size):
        """
        create batch of document embeddings
        """

        for i in range(0, len(documents), embeddings_batch_size):
            yield documents[i: i + embeddings_batch_size]

    def create_db_from_batch_documents(self, documents, embeddings_model):
        """
        Create DB from documents in batches, this is used for chromadb to get around the add limit
        in sqlite implementation of later versions of chromadb
        """
        batches_documents = self._create_batch_embeddings(
            documents, self.args.embeddings_batch_size
        )
        n_batches = len(documents) // self.args.embeddings_batch_size \
            if len(documents) >= self.args.embeddings_batch_size else 1
        try:
            for batch_id, batch_document in enumerate(batches_documents):
                _ = self.create_db_from_documents(batch_document, embeddings_model)
                logger.info(f"Added batch {batch_id + 1} of {n_batches} batches to vector db")

        except Exception as e:
            raise Exception(
                f"Error loading embeddings batches to {self.args.vector_store_name}: {e}"
            )

    def embeddings_to_vectordb(self) -> None:
        """Create vectorstore from documents and store locally."""

        start_time = time.time()

        # Load documents and splits in chunks (if not in evaluation_type mode)
        documents = self.split_documents(
            chunk_size=self.args.chunk_size, chunk_overlap=self.args.chunk_overlap
        )

        # Load embeddings model
        embeddings_model = load_embeddings_model(
            self.embeddings_model_name, self.args.use_gpu
        )

        logger.info(f"Creating vectorstore from documents")

        if not validate_documents(documents):
            raise ValueError("Invalid documents")

        if self.args.vector_store_name == "chromadb":

            # chromadb does autosave in latest version, just for the saver PersistedVectorStoreSaver
            db = self.create_db_from_batch_documents(documents, embeddings_model)

        else:
            try:
                db = self.create_db_from_documents(documents, embeddings_model)
            except Exception as e:
                raise Exception(
                    f"Error loading embeddings to {self.args.vector_store_name}: {e}"
                )

        config = PersistedVectorStoreSaverConfig(
            vector_store_name=self.args.vector_store_name,
            vector_store=db,
            persist_directory=self.args.vector_store_storage_path,
            collection_name=self.args.collection_name,
        )

        vector_store_saver = PersistedVectorStoreSaver(config)

        vector_store_saver.save_vector_store(db)

        db = None  # Free up memory

        end_time = time.time()
        elapsed_time = round(end_time - start_time)

        logger.info(f"Finished creating {self.args.vector_store_name} from texts, it has been "
                    f"persisted to {self.args.vector_store_storage_path}")

        time_minutes = round(elapsed_time / 60)

        if time_minutes > 1:
            logger.info(f"Elapsed time: {time_minutes} minutes")
        else:
            logger.info(f"Elapsed time: {elapsed_time} seconds")
